"""
Base object for scheduling
"""

from __future__ import annotations

from moat.util import load_ext


class BaseLoader:
    """
    Describe how to load various scheduling data.
    """

    @staticmethod
    async def price_buy(cfg, t):
        """
        Future prices for incoming energy, in $$$/kWh.
        """
        raise NotImplementedError("You need to add a static 'price_buy' method")

    @staticmethod
    async def price_sell(cfg, t):
        """
        Future prices for sold energy, in $$$/kWh.
        """
        raise NotImplementedError("You need to add a static 'price_sell' method")

    @staticmethod
    async def solar(cfg, t):
        """
        Projected solar power, in kW.
        """
        raise NotImplementedError("You need to add a static 'solar' method")

    @staticmethod
    async def load(cfg, t):
        """
        Projected local consumption, in kW.
        """
        raise NotImplementedError("You need to add a static 'load' method")

    @staticmethod
    async def soc(cfg):
        """
        The battery's current SoC.
        """
        res = cfg.bms.sched.start.soc
        if res < 0:
            raise NotImplementedError("You need to set the 'start.soc' parameter")
        return res

    @staticmethod
    async def result(cfg, **kw):
        """
        Accepts the intended state at the end of the first period.
        power to/from the grid and intended SoC.
        """
        raise NotImplementedError("You need to add a static 'result' method")

    @staticmethod
    async def results(cfg, it):
        """
        Accepts an iterator for intended results over all periods.

        By default this does nothing.
        """
        pass  # pylint:disable=unnecessary-pass


def Loader(name, key=None):
    """Fetch a named loader class"""
    res = load_ext(f"moat.bms.sched.mode.{name}")
    if res is None:
        raise AttributeError(name)
    res = res.Loader
    if key is not None:
        res = getattr(res, key)
    return res



"""
Germany: Get next-day prices from awattar.de API
"""

from __future__ import annotations

import asks

from . import BaseLoader


class Loader(BaseLoader):
    """
    Next-day spot market prices for Germany.

    Source: Awattar.
    """

    @staticmethod
    async def price_sell(cfg, t):
        """
        Read prices for the German next-day spot market from the Awattar API.
        """
        start = t - 24 * 3600
        end = start + 7 * 24 * 3600
        factor = cfg.data.awattar.factor
        offset = cfg.data.awattar.offset

        async with asks.sessions.Session() as s:
            r = await s.get(cfg.data.awattar.url, params=dict(start=start * 1000, end=end * 1000))
            dd = r.json()["data"]
            for d in dd[24:]:
                val = d["marketprice"] * factor + offset
                for _ in range(cfg.steps):
                    yield val

        # Somewhat primitive linear interpolation from past data
        for _ in range(cfg.data.awattar.extend):
            s = len(dd) - 24
            dv = dd[-25]["marketplace"] - dd[-1]["marketplace"]
            for i in range(24):
                val = (dd[-24 + i]["marketplace"] - dv * (25 - i) / 25) * factor + offset
                for _ in range(cfg.steps):
                    yield val



#!/usr/bin/env python3
"""
Basic tool support

"""

from __future__ import annotations

import logging
import time
from textwrap import dedent as _dedent

import asyncclick as click
from moat.util import (
    attr_args,
    list_ext,
    load_ext,
    load_subgroup,
    merge,
    process_args,
    yload,
    yprint,
)

from .control import Model
from .mode import BaseLoader

log = logging.getLogger()


def dedent(s):
    "remove indent"
    return _dedent(s).strip()


def Loader(name, key=None):
    """load a named mode"""
    res = load_ext(f"moat.bms.sched.mode.{name}")
    if key is False:
        return res
    res = res.Loader
    if key is not None:
        res = getattr(res, key)
    return res


@load_subgroup(prefix="moat.bms.sched")
@click.pass_obj
@click.option(
    "-c",
    "--config",
    help="Configuration file (YAML)",
    type=click.Path(dir_okay=False, readable=True),
)
@attr_args(with_path=False, with_proxy=False)
async def cli(obj, config, **attrs):
    """Battery Manager: Scheduling"""

    cfg = obj.cfg.bms.sched
    if config:
        with open(config) as f:
            cc = yload(f)
            merge(cfg, cc)
    obj.cfg.bms.sched = process_args(cfg, **attrs)


@cli.command()
@click.pass_obj
def dump(obj):
    """
    Dump the current configuration as YAML
    """
    yprint(obj.cfg.bms.sched)


@cli.command()
@click.argument("name", nargs=-1)
def modes(name):
    """List known modes / help text for a mode"""
    static = dict(
        T="""\
List of known inputs+outputs. Use T.‹name› or ‹mode›.‹name› for details.
""",
    )
    if not name:
        mn = [
            # pylint:disable=unexpected-keyword-arg
            m
            for m, _ in list_ext("moat.bms.sched.mode", pkg_only=False)
        ]
        mn.extend(static.keys())
        ml = max(len(m) for m in mn)
        for m in mn:
            if m in static:
                doc = static[m]
            else:
                try:
                    mm = Loader(m)
                except ImportError as exc:
                    doc = repr(exc)
                else:
                    doc = dedent(mm.__doc__).split("\n", 1)[0]
            print(f"{m:{ml}s}  {doc}")
        return
    for m in name:
        if m == "T":
            print("T:")
            ml = max(len(x) for x in dir(BaseLoader) if not x.startswith("_"))
            for m in dir(BaseLoader):
                if m.startswith("_"):
                    continue
                doc = dedent(getattr(BaseLoader, m).__doc__).split("\n", 1)[0]
                print(f"{m:{ml}s}  {doc}")
            continue
        if m.startswith("T."):
            doc = dedent(getattr(BaseLoader, m[2:]).__doc__)
        elif "." in m:
            mn, *a = m.split(".")
            mo = Loader(mn)
            for aa in a:
                mo = getattr(mo, aa)
            doc = dedent(mo.__doc__)
        else:
            m = Loader(m)
            doc = dedent(m.__doc__)
            doc += "\nImplements: " + " ".join(x for x in m.__dict__ if not x.startswith("_"))

        print(
            f"""\
{m}:
{doc}
""",
        )


@cli.command(
    help="""
Calculate proposed SoC by analyzing files with assumed future usage and weather / solar input.
Goal: minimize cost.
""",
)
@click.pass_obj
@click.option(
    "-a",
    "--all",
    "all_",
    is_flag=True,
    help="emit all outputs (default: first interval)",
)
@click.option("-f", "--force", is_flag=True, help="Run even if we're not close to the timeslot")
async def analyze(obj, all_, force):
    """
    Analyze future data.
    """
    cfg = obj.cfg.bms.sched

    t = None
    if force:
        t_slot = 3600 / cfg.steps
        t = time.time() + t_slot / 2
        t -= t % t_slot

    m = Model(cfg, t)

    soc_cur = cfg.start.soc
    if soc_cur < 0:
        soc_fn = cfg.mode.soc
        if soc_fn is None:
            raise click.ClickException("I need to know the current SoC")
        soc_cur = await Loader(soc_fn, "soc")(cfg)

    if all_:
        cfg.mode.result = None
        cfg.mode.results = "file"

    await m.propose(soc_cur)



"""
Get solar data from forecast.solar
"""

from __future__ import annotations

from datetime import datetime

import asks

from . import BaseLoader


class Loader(BaseLoader):
    """
    Get solar yield forceat from the "forecast.solar" API.
    """

    @staticmethod
    async def _solar(cfg, t, session, a=None):
        """
        Projected solar power, in kW.

        Config:
            data.fore_solar.url: base URL
            data.fore_solar.api: your API key
            data.fore_solar.factor: 0.001
        """

        factor = cfg.data.fore_solar.factor
        start = datetime.utcfromtimestamp(t - 3600).strftime("%H:%M")
        t_step = int(3600 / cfg.steps)
        cmp = a["compass"]
        if cmp > 180:
            cmp -= 360

        url = (
            f"{cfg.data.fore_solar.url}/{cfg.data.fore_solar.api}/estimate/"
            + f"watts/{cfg.solar.lat}/{cfg.solar.long}/{a['tilt']}/{cmp}/{int(a['peak'] * 1000)}"
        )
        r = await session.get(
            url,
            headers=dict(
                accept="application/json",
            ),
            params=dict(
                no_sun=0,
                time="seconds",
                start=start,
                damping=cfg.solar.damping,
            ),
        )
        if r.status_code >= 400:
            raise RuntimeError(r, r.content)
        kx = 0
        tn = t + t_step
        vs = 0  # value sum
        n = 0  # #values

        # If the result set is more detailed than our request, we
        # average the data.
        # If the result is less detailed, we repeat values.
        #
        # XXX there's no interpolation, thus if the result has a
        # half-hour resolution but our step size is 3/hour, the result
        # is somewhat less accurate than theoretically possible. In
        # practice this should not matter too much..
        for k, v in r.json()["result"].items():
            k = int(k)
            # must be ascending
            assert k > kx
            kx = k

            if k < t:
                continue
            if k < tn:
                vs += v
                n += 1
                continue

            while k >= tn:
                t = tn
                tn += t_step

                yield vs / 1000 / n * factor
                # value is kW, thus not divided by t_step
                vs = v
                n = 1

        if n:
            for _ in range(cfg.steps):
                yield vs / 1000 / n * factor

    @classmethod
    async def solar(cls, cfg, t):
        async with asks.sessions.Session(connections=2) as s:
            a = [cls._solar(cfg, t, s, a) for a in cfg.solar.array]
            while True:
                val = 0
                for v in a:
                    try:
                        val += await anext(v)
                    except StopAsyncIteration:
                        return
                yield val



"""
Read data from a file
"""

from __future__ import annotations

import sys

import anyio

from . import BaseLoader


class Loader(BaseLoader):
    """
    Load scheduling data from a file.
    """

    @staticmethod
    async def _file(cfg, key):
        async with await anyio.Path(getattr(cfg.data.file, key)).open("r") as f:
            async for line in f:
                yield float(line.strip())

    @staticmethod
    async def price_buy(cfg, t):
        """
        Read future prices for incoming energy, in $$$/kWh.
        File format: one float per line.

        Config:
            data.file.price_buy: path to the data file.
        """
        async for x in Loader._file(cfg, "price_buy"):
            yield float(x)

    @staticmethod
    async def price_sell(cfg, t):
        """
        Read future prices for sold energy, in $$$/kWh.
        File format: one float per line.

        Config:
            data.file.price_sell: path to the data file.
        """
        async for x in Loader._file(cfg, "price_sell"):
            yield float(x)

    @staticmethod
    async def solar(cfg, t):
        """
        Projected solar power, in kW.
        File format: one float per line.

        Config:
            data.file.solar: path to the data file.
        """
        async for x in Loader._file(cfg, "solar"):
            yield float(x)

    @staticmethod
    async def load(cfg, t):
        """
        Projected local consumption, in kW.
        File format: one float per line.

        Config:
            data.file.load: path to the data file.
        """
        async for x in Loader._file(cfg, "load"):
            yield float(x)

    @staticmethod
    async def soc(cfg):
        """
        The battery's SoC at the start of a simulation.

        File format: one float.

        Config:
            data.file.soc: path to the file.
        """
        async for x in Loader._file(cfg, "soc"):
            return float(x)

    @staticmethod
    async def result(cfg, **kw):
        """
        Send the result data to a file.

        Config:
            data.result.format: yaml, json, or msgpack
        """
        f = cfg.data.format.result
        if f == "yaml":
            from moat.util import yformat

            async with await anyio.Path(cfg.data.file.result).open("w") as f:
                await f.write(yformat(kw))
        elif f == "msgpack":
            from moat.util import packer

            async with await anyio.Path(cfg.data.file.result).open("wb") as f:
                await f.write(packer(kw))
        elif f == "json":
            import json

            async with await anyio.Path(cfg.data.file.result).open("w") as f:
                await f.write(json.dumps(kw))
        else:
            print(f"Unknown output format {f!r}. Use yaml/msgpack/json.")
            sys.exit(1)

    @staticmethod
    async def results(cfg, it):
        """
        Print the resulting data, as a YAML array.

        Config:
            data.results.format: yaml, json, or msgpack
        """
        f = cfg.data.format.results
        if f == "yaml":
            from moat.util import yformat

            async with await anyio.Path(cfg.data.file.result).open("w") as f:
                async for kw in it:
                    await f.write(yformat([kw]))
        elif f == "msgpack":
            res = []
            async for kw in it:
                res.append(kw)
            from moat.util import packer

            async with await anyio.Path(cfg.data.file.result).open("wb") as f:
                await f.write(packer(res))
        elif f == "json":
            import json

            async for kw in it:
                res.append(kw)
            async with await anyio.Path(cfg.data.file.result).open("w") as f:
                await f.write(json.dumps(res))
        else:
            print(f"Unknown output format {f!r}. Use yaml/msgpack/json.")
            sys.exit(1)



"""
Derive buy from sell price
"""

from __future__ import annotations

from . import BaseLoader
from . import Loader as _Loader


class Loader(BaseLoader):
    """
    Derive buy from sell price.

    This importer calculates the cost for buying energy from those of sold energy.
    """

    @staticmethod
    async def price_buy(cfg, t):
        """
        Read future prices for incoming energy, in $$$/kWh.
        File format: one float per line.

        Config:
            data.file2.const (float): additional fixed price per kWh.
            data.file2.factor (float): Multiplicator for the values read from the file.
        """
        factor = cfg.data.file2.factor
        offset = cfg.data.file2.offset

        async for x in _Loader(cfg.mode.price_sell).price_sell(cfg, t):
            yield float(x) * factor + offset



"""
Charge/discharge optimizer.
"""

from __future__ import annotations

import datetime
import logging
import time
from contextlib import nullcontext

import anyio
from moat.util import attrdict
from moat.util.times import humandelta, ts2iso
from ortools.linear_solver import pywraplp

from .mode import Loader

logger = logging.getLogger(__name__)


async def generate_data(cfg, t):
    # loads, prices, solar, buy_factor, buy_const):
    """
    Generate data chunks from iterators
    """
    iters = {}
    for k in "price_buy,price_sell,solar,load".split(","):
        iters[k] = aiter(Loader(cfg.mode[k], k)(cfg, t))
    while True:
        val = attrdict()
        k = None
        try:
            for k, v in iters.items():
                val[k] = await anext(v)
        except StopAsyncIteration:
            logger.info("END: %s", k)
            return
        logger.debug("DATA: %r", val)
        yield val


def add_piecewise(solver, x, y, points: list[int, int], name):
    """
    Add a piecewise-linear constraint on y=a*x+b.

    @points is a list of x/y pairs to interpolate between.
    """

    # Source:
    # https://or.stackexchange.com/questions/6674/how-to-linearize-specific-range-constraints#answer-6675

    # untested

    n = len(points) - 1  # number of segments
    l = []  # noqa:E741  # lambda values to interpolate within a segment

    rx = []
    ry = []
    for i, xyi in enumerate(points):
        xi, yi = xyi
        li = solver.NumVar(0, 1, f"l_{name}_{i}")

        l.append(li)
        rx.append(xi * li)
        ry.append(yi * li)

    b = [solver.BoolVar(f"b_{name}_{i}") for i in range(n)]
    solver.Add(sum(l) == 1)  # 1: lambda values sum to 1
    solver.Add(sum(b) == 1)  # 3: we're in exactly one segment
    solver.Add(x == sum(rx))  # 2: interpolated X value
    solver.Add(y == sum(ry))  # Objective: interpolated Y value
    for i in range(n):
        solver.Add(b[i] <= l[i] + l[i + 1])
        # 4-7: when x is in segment i, the sum of the lambdas on
        # the upper and lower boundary of this segment is forced to be at
        # least 1. Because of (1) and the fact that lambda is constrained
        # to [0,1], this sum must be exactly 1 and all other lambda values
        # are thus zero.


class Model:
    """
    Calculate optimum charge/discharge behavior based on
    minimizing cost / maximizing income.

    Initial input:
    * configuration data
    * async iterator for values

    Solver input:
    * current charge level (SoC)

    Output (first period):
    * grid input (negative: send energy)
    * battery SoC at end of first period

    You should re-run the model at the start of the next period, using
    the SoC at that time.
    """

    money = None
    cap = None
    g_sell = None
    g_buy = None
    g_sells = None
    g_buys = None
    moneys = None
    caps = None
    b_diss = None
    b_chgs = None
    cap_init = None
    constr_init = None
    objective = None
    solver = None

    def __init__(self, cfg: dict, t=None):
        if t is None:
            t_slot = 3600 / cfg.steps
            t_now = time.time()
            t = t_now + t_slot / 2
            t -= t % t_slot
            if abs(t_now - t) > t_slot / 10:
                raise ValueError(
                    f"You're {humandelta(abs(t_now - t))} away "
                    f"from {ts2iso(t)}",
                )

        elif t % (3600 / cfg.steps):
            raise ValueError(f"Time {t} not a multiple of 3600/{cfg.steps}")
        self.cfg = cfg
        self.t = t

    async def _setup(self):
        cfg = self.cfg
        data = generate_data(cfg, self.t)
        per_hour = self.cfg.steps

        # ORtools
        self.solver = solver = pywraplp.Solver("B", pywraplp.Solver.GLOP_LINEAR_PROGRAMMING)
        self.objective = solver.Objective()
        inf = solver.infinity()

        # Starting battery charge
        self.cap_init = cap_prev = solver.NumVar(
            cfg.battery.capacity * 0.05,
            cfg.battery.capacity * 0.95,
            "b_init",
        )
        self.constr_init = solver.Constraint(0, 0)
        self.constr_init.SetCoefficient(self.cap_init, 1)

        # collect vars for reporting
        self.g_buys = []
        self.g_sells = []
        self.b_chgs = []
        self.b_diss = []
        self.caps = []
        self.moneys = []

        i = 0
        _pr = None
        while True:
            # input constraints
            try:
                dt = await anext(data)
            except StopAsyncIteration:
                break
            if dt.price_buy == dt.price_sell:
                dt.price_buy *= 1.001
            elif dt.price_buy < dt.price_sell:
                raise ValueError(f"At {i}: buy {dt.price_buy} < sell {dt.price_sell} ??")
                # TODO

            # ### Variables to consider

            # future battery charge
            cap = solver.NumVar(
                cfg.battery.capacity * cfg.battery.soc.min,
                cfg.battery.capacity * cfg.battery.soc.max,
                f"b{i}",
            )

            self.caps.append(cap)

            # battery charge/discharge
            b_chg = solver.NumVar(0, cfg.battery.max.charge / per_hour, f"bc{i}")
            b_dis = solver.NumVar(0, cfg.battery.max.discharge / per_hour, f"bd{i}")
            self.b_chgs.append(b_chg)
            self.b_diss.append(b_dis)

            # solar power input. We may not be able to take all
            s_in = solver.NumVar(0, dt.solar / per_hour, f"pv{i}")

            # inverter charge/discharge
            i_chg = solver.NumVar(0, cfg.inverter.max.charge / per_hour, f"ic{i}")
            i_dis = solver.NumVar(0, cfg.inverter.max.discharge / per_hour, f"id{i}")

            # local load
            l_out = solver.NumVar(dt.load / per_hour, dt.load / per_hour, f"ld{i}")

            # grid
            g_buy = solver.NumVar(0, cfg.grid.max.buy / per_hour, f"gi{i}")
            g_sell = solver.NumVar(0, cfg.grid.max.sell / per_hour, f"go{i}")
            self.g_buys.append(g_buy)
            self.g_sells.append(g_sell)

            # income to maximize
            money = solver.NumVar(-inf, inf, f"pr{i}")
            self.moneys.append(money)

            # ### Constraints (actually Relationships, as they're all equalities)

            # Battery charge. old + charge - discharge == new, so … - new == 0.
            solver.Add(cap_prev + cfg.battery.efficiency.charge * b_chg - b_dis == cap)

            # DC power bar. power_in == power_out
            solver.Add(
                s_in
                + cfg.battery.efficiency.discharge * b_dis
                + cfg.inverter.efficiency.charge * i_chg
                == b_chg + i_dis,
            )

            # AC power bar. power_in == power_out
            solver.Add(g_buy + cfg.inverter.efficiency.discharge * i_dis == g_sell + l_out + i_chg)

            # Money earned: grid_out*price_sell - grid_in*price_buy
            solver.Add(
                dt.price_sell * g_sell
                - dt.price_buy * g_buy
                + cap
                * cfg.battery.soc.value.current
                / cfg.battery.capacity  # bias for keeping the battery charged
                == money,
            )

            self.objective.SetCoefficient(money, 1)
            cap_prev = cap
            if not i:
                self.g_buy, self.g_sell = g_buy, g_sell
                self.cap = cap
                self.money = money
            i += 1

        # Attribute a fake monetary value of ending up with a charged battery
        self.objective.SetCoefficient(cap, cfg.battery.soc.value.end / cfg.battery.capacity)

        self.objective.SetMaximization()

    async def propose(self, charge):
        """
        Assuming that the current SoC is @charge, return
        - how much power to take from / -feed to the grid [W]
        - the SoC at the end of the current period [0…1]
        - this period's earnings / -cost [$$]

        """
        await self._setup()
        cfg = self.cfg

        charge *= cfg.battery.capacity
        self.constr_init.SetLb(charge)
        self.constr_init.SetUb(charge)

        self.solver.Solve()

        async with anyio.create_task_group() as tg:
            res = cfg.mode.result
            if res is not None:
                res = Loader(res).result

            res2 = cfg.mode.results
            if res2 is not None:
                sch, rch = anyio.create_memory_object_stream(1)
                tg.start_soon(Loader(res2).results, cfg, rch)
                res2 = sch

            async with res2 if res2 is not None else nullcontext():
                for g_buy, g_sell, b_chg, b_dis, cap, money in zip(
                    self.g_buys,
                    self.g_sells,
                    self.b_chgs,
                    self.b_diss,
                    self.caps,
                    self.moneys,
                    strict=False,
                ):
                    val = dict(
                        grid=(g_buy.solution_value() - g_sell.solution_value()) * cfg.steps,
                        soc=cap.solution_value() / cfg.battery.capacity,
                        batt=b_chg.solution_value() - b_dis.solution_value(),
                        money=money.solution_value(),
                    )

                    if res is not None:
                        tg.start_soon(res, cfg, val)
                        res = None
                        if res2 is None:
                            break

                    if res2 is not None:
                        await res2.send(val)
